module Parsing.LineParsing where

import Parsing.Syntax
import Parsing.SrcLoc

import Common.Tree
import Common.Utils

import Process.Result

import Data.Map hiding (map)
import Text.Regex.Posix
import qualified List



----------------------------------------------------------------------------------
--	Result types
----------------------------------------------------------------------------------

type LineParser  = Stream -> Result ParseTree ParseError
type MultiParser = Stream -> Result [ParseTree] ParseError

type ParseError = (SrcLoc, String)
type ParseTree = Tree ParseNode


{---------------------------------------------------------------------------------
--	Parse nodes & constructor helpers
---------------------------------------------------------------------------------}

data ParseNode
	= SyntaxNode {
		name :: String,
		start :: SrcLoc,
		end :: SrcLoc
		}
	| TokenNode {
		name :: String,
		start :: SrcLoc,
		end :: SrcLoc,
		value :: String
		}
	| EmptyNode

instance Show ParseNode where
	show (SyntaxNode name s e) = show s ++ "-" ++ show e ++ ": " ++ name
	show (TokenNode name s e val) = show s ++ "-" ++ show e ++ ": " ++ name ++ " " ++ val
	show (EmptyNode) = ""

tokenNode :: SrcLoc -> String -> String -> ParseTree
tokenNode loc name token = Tree {
	children = [],
	content = TokenNode {
		name = name,
		start = loc,
		end = loc `shift` (length token - 1),
		value = token
		}}

syntaxNode :: String -> [ParseTree] -> ParseTree
syntaxNode name elems = Tree {
	content = SyntaxNode {
		name = name,
		start = (start . content . head) elems,
		end   = (end   . content . last) elems
		},
	children = elems
	}

emptyNode = Tree EmptyNode []


-----------------------------------------------------------------------------
--	Character stream level utility functions
-----------------------------------------------------------------------------

after :: Stream -> ParseTree -> Stream
after (loc, str) tree = (loc', str') where
	loc' = (end (content tree)) `shift` 1
	str' = drop (char loc' - char loc) str


matches :: String -> String -> Bool
matches str pattern = str =~ ("^(" ++ pattern ++ ")$")

grab :: String -> String -> String
grab pattern str = str =~ ("^(" ++ pattern ++ ")")


startOf :: String -> String
startOf str 
	| length str <= 10 = str
	| otherwise        = (take 10 str) ++ "..."


-- Skips blanks and comments
makeSkip :: SyntaxGrammar -> (Stream -> Stream)
makeSkip syntax = skip where
	skip stream = consume taken stream where
		taken = grab skipRegex (str stream)
		skipRegex = concat $ ["(("] ++ List.intersperse ")|(" (skipped syntax)  ++ ["))*"]


-----------------------------------------------------------------------------
--	Creating the parsing functions
-----------------------------------------------------------------------------

makeRuleParers :: SyntaxGrammar -> (String -> LineParser)
makeRuleParers syntax = parseRule where
	
	-- All rule parsers
	ruleParsers :: Data.Map.Map String LineParser
	ruleParsers = fromList $ map ruleParser (rules syntax)
	
	-- Selects the right parser
	parseRule :: String -> LineParser
	parseRule name = case Data.Map.lookup name ruleParsers of
		Just p -> p
		Nothing -> (\ s -> Failure (loc s, "The syntax grammar does not contain a rule named '" ++ name ++ "'."))
	
	
	-----------------------------------------------------------------------------
	--	Utility functions
	-----------------------------------------------------------------------------
	
	-- TODO: should use regexpt instead
	isBoundary :: Char -> Bool
	isBoundary c = [c] `elem` (delimiters syntax) || [c] `elem` (skipped syntax)
	
	-- Skips blanks and comments
	skip :: Stream -> Stream
	skip = makeSkip syntax
	
	-----------------------------------------------------------------------------
	--	Creating a rule parser
	-----------------------------------------------------------------------------
	ruleParser :: (String, ParseRule) -> (String, LineParser)
	
	-----------------------------------------------------------------------------
	--	Creating a token parser
	-----------------------------------------------------------------------------
	ruleParser (name, Token ttype regexp) = (name, parser) where
		parser :: LineParser
		parser (loc, "")  = Failure (loc, "end of line found but " ++ name ++ " expected")
		parser (loc, str) = case ttype of
			Bounded -> if token `matches` regexp
				then Success tokenNode'
				else Failure (loc, token ++ " found but " ++ name ++ " expected")
			Greedy -> if token /= ""
				then if isBounded
					then Success tokenNode'
					else Failure (loc `shift` (length token), "delimiter or white space expected after " ++ token)
				else Failure (loc, "failed to parse " ++ name)
			where
				boundedToken :: String
				boundedToken = if [head str] `elem` (delimiters syntax)
					then [head str]
					else takeWhile (not . isBoundary) str
				
				greedyToken :: String
				greedyToken = grab regexp str
	
				token :: String
				token = case ttype of
					Bounded -> boundedToken
					Greedy  -> greedyToken
				
				isBounded :: Bool
				isBounded = rem == "" || (isBoundary . head) rem where
					rem = drop (length token) str
				
				tokenNode' = tokenNode loc name token
	
	
	
	-----------------------------------------------------------------------------
	--	Creating a syntax rule parser
	-----------------------------------------------------------------------------
	ruleParser (name, Syntax rule) = (name, parser) where
		parser stream = case (ruleContentParser rule) stream of
			Success elems -> Success (syntaxNode name elems)
			Failure f  -> Failure (furthest defErr f)
			where
				defErr = (loc stream, name ++ " expected")
		
		
		-----------------------------------------------------------------------------
		--	Rule specification -> parse function parsing the rule content
		--	The output of the rule content parsers is a list of nodes
		-----------------------------------------------------------------------------	
		ruleContentParser :: SyntaxRule -> MultiParser

		ruleContentParser (Either rules) = parser where
			subParsers = map ruleContentParser rules
			parser (loc, str) = alt subParsers defErr where
				defErr = (loc, "Unexpected characters: \"" ++ startOf str ++ "\"")
				alt [] err = Failure err
				alt (p:ps) err = case p (loc, str) of
					Success s -> Success s
					Failure f -> alt ps (furthest err f)
		
			
		ruleContentParser (Sequence rules) = parser where
			subParsers = map ruleContentParser rules
			parser stream = seq subParsers [] where
				seq [] parsed = Success parsed
				seq (p:ps) parsed = case p rem of
					Success s -> seq ps (parsed ++ s)
					Failure f -> Failure f
					where
						rem = case parsed of
							[] -> stream
							_  -> skip $ stream `after` (last parsed)
		
		
		ruleContentParser (Maybe rule) = parser where
			parser stream = case (ruleContentParser rule) stream of
				Failure f -> Success []
				Success s -> Success s
		
	
		ruleContentParser (ZeroOrMore rule) = parser where
			subParser = ruleContentParser rule
			parser stream = zom stream []
			zom stream parsed = case subParser stream of
				Success s -> continueZom s
				Failure f -> Success parsed
				where
					continueZom s = zom stream' parsed' where
						stream' = case parsed' of
							[] -> stream
							_  -> skip $ stream `after` (last parsed')
						parsed' = parsed ++ s
		
		
		ruleContentParser (OneOrMore rule) = ruleContentParser (Sequence [rule, ZeroOrMore rule])
		
		
		ruleContentParser (Rule name) = parser where
			parser stream = case (parseRule name) stream of
				Success node -> Success [node]
				Failure f    -> Failure f
	
	
		ruleContentParser (DirectToken tok) = parser where
			parser (loc, str) = if tok `List.isPrefixOf` str
				then Success [tokenNode loc "" tok]
				else Failure (loc, tok ++ " expected.")
					
				
		dummyParsing (loc, str) = Failure (loc, "Not implemented, TODO.")

